#!/usr/bin/env python3

### IMPORTS ###
import logging
import importlib

from kneedeepio.plugins.plugin import Plugin

from .exceptions import ServiceAlreadyRegistered
from .exceptions import ServiceNotRegistered
from .exceptions import PluginAlreadyLoadedException
from .exceptions import PluginNotLoadedException

### GLOBALS ###

### FUNCTIONS ###

### CLASSES ###
class PluginFactory:
    def __init__(self, logging_srv):
        self.logger = logging.getLogger(type(self).__name__)
        self.logger.debug("Inputs - logging_srv: %s", logging_srv)

        self._service_registry = {}
        self._plugin_registry = []
        self._load_callbacks = []
        self._unload_callbacks = []

        self.register_service("logging", logging_srv)

    def register_service(self, service_type, service):
        self.logger.debug("Inputs - service_type: %s, service: %s", service_type, service)
        # Check if service type already registered.
        if service_type in self._service_registry:
            raise ServiceAlreadyRegistered("Service of type {} already registered.".format(service_type))
        # Put service in registry
        self._service_registry[service_type] = service

    def load(self, module_name, class_name):
        self.logger.debug("Inputs - module_name: %s, class_name: %s", module_name, class_name)
        # Check if plugin is already loaded
        for loaded_plugin in self._plugin_registry:
            if loaded_plugin["module_name"] == module_name and loaded_plugin["class_name"] == class_name:
                raise PluginAlreadyLoadedException
        # Import the plugin
        tmp_module = importlib.import_module(module_name)
        self.logger.debug("tmp_module: %s", tmp_module)
        # Create an instance of the plugin
        tmp_class = getattr(tmp_module, class_name)
        if not issubclass(tmp_class, Plugin):
            raise TypeError("Plugin does not subclass kneedeepio.plugins.plugin.Plugin")
        self.logger.debug("tmp_class: %s", tmp_class)
        # NOTE: The logging service is always provided as it should always be used.
        tmp_services = {"logging": self._service_registry["logging"]}
        for tmp_service_type in tmp_class.required_services:
            if tmp_service_type in self._service_registry:
                tmp_services[tmp_service_type] = self._service_registry[tmp_service_type]
            else:
                raise ServiceNotRegistered("Service type '{}' not registered.".format(tmp_service_type))
        tmp_instance = tmp_class(tmp_services)
        self.logger.debug("tmp_instance: %s", tmp_instance)
        # Store the instance in the registry list
        self._plugin_registry.append({
            "module_name": module_name,
            "class_name": class_name,
            "instance": tmp_instance
        })
        # Run the plugin instance setup
        tmp_instance.setup()
        # Call the load callbacks
        for cb in self._load_callbacks:
            cb(tmp_instance)

    def unload(self, module_name, class_name):
        self.logger.debug("Inputs - module_name: %s, class_name: %s", module_name, class_name)
        # Check if plugin is already loaded
        tmp_plugin = None
        for loaded_plugin in self._plugin_registry:
            if loaded_plugin["module_name"] == module_name and loaded_plugin["class_name"] == class_name:
                tmp_plugin = loaded_plugin
        if tmp_plugin is None:
            raise PluginNotLoadedException
        # Call the unload callbacks
        for cb in self._unload_callbacks:
            cb(tmp_plugin["instance"])
        # Run the plugin instance teardown
        tmp_plugin["instance"].teardown()
        # Remove the instance from the registry
        self._plugin_registry.remove(tmp_plugin)
        # FIXME: How to un-import the plugin module?
        #        Is the un-import necessary?
        #        Have to check to make sure there aren't any other classes using
        #           the same module.

    def register_load_callback(self, callback_method):
        self.logger.debug("Inputs - callback_method: %s", callback_method)
        # Add the callback method to the list of methods to call back on plugin load.
        # The callback method should take one argument: the instance of the plugin.
        self._load_callbacks.append(callback_method)

    def register_unload_callback(self, callback_method):
        self.logger.debug("Inputs - callback_method: %s", callback_method)
        # Add the callback method to the list of methods to call back on plugin unload.
        # The callback method should take one argument: the instance of the plugin.
        self._unload_callbacks.append(callback_method)

    def tick_plugins(self):
        self.logger.debug("Inputs - None")
        # Call the tick function for each of the plugins.
        # This can be used as a heartbeat for the plugin, or used to perform a
        #    small amount of work.
        for item in self._plugin_registry:
            item["instance"].tick()


#!/usr/bin/env python3

### IMPORTS ###
import abc

from kneedeepio.plugins.services import LoggingService

### GLOBALS ###

### FUNCTIONS ###

### CLASSES ###
class Plugin(metaclass = abc.ABCMeta):
    required_services = ["logging"] # NOTE: The logging service is always supplied, but it's good to be explicit.
    def __init__(self, services):
        # for service in services:
        #     if isinstance(services[service], ConfigurationService):
        #         self.config = services[service]
        #     # else:
        #     #     raise TypeError("Invalid object passed as Configuration Service")
        #
        #     if isinstance(services[service], LoggingService):
        #         self.logger = services[service]
        #     # else:
        #     #     raise TypeError("Invalid object passed as Loggin Service")
        #
        #     # if isinstance(metrics_srv, MetricsReportingService):
        #     #     self.metric = metrics_srv
        #     # else:
        #     #     raise TypeError("Invalid object passed as Metrics Reporting Service")
        #
        #     if isinstance(services[service], ObjectDatastoreService):
        #         self.object = services[service]
        #     # else:
        #     #     raise TypeError("Invalid object passed as Object Datastore Service")
        if isinstance(services["logging"], LoggingService):
            self.logger = services["logging"]
        else:
            raise TypeError("Invalid object passed as Logging Service")

        self.logger.debug("Plugin Initialized: %s", type(self).__name__)

    @abc.abstractmethod
    def setup(self):
        # This method should do whatever is needed to initialize the plugin's operation.
        raise NotImplementedError

    @abc.abstractmethod
    def tick(self):
        # FIXME: Pick a better name for this
        # This method should do some small piece of work for the plugin.
        raise NotImplementedError

    @abc.abstractmethod
    def teardown(self):
        # This method should do whatever is needed to un-setup the plugin, stopping operation.
        raise NotImplementedError



#!/usr/bin/env python3

### IMPORTS ###
import abc
import logging

### GLOBALS ###

### FUNCTIONS ###

### CLASSES ###
class ConfigurationService(metaclass = abc.ABCMeta):
    def __init__(self):
        self.logger = logging.getLogger(type(self).__name__)

    @abc.abstractmethod
    def load_config(self, config):
        raise NotImplementedError

    @abc.abstractmethod
    def get_value(self, key):
        raise NotImplementedError

    @abc.abstractmethod
    def set_value(self, key, value):
        raise NotImplementedError

class InMemoryConfigurationService(ConfigurationService):
    def __init__(self):
        super().__init__()
        self.logger.debug("Inputs - None")
        self._data = {}

    def load_config(self, config):
        self.logger.debug("Loading config: %s", config)
        for item_key, item_value in config.items():
            self.set_value(item_key, item_value)

    def get_value(self, key):
        self.logger.debug("Getting value for key: %s - %s", key, self._data[key])
        return self._data[key]

    def set_value(self, key, value):
        self.logger.debug("Setting value for key: %s - %s", key, value)
        self._data[key] = value


#!/usr/bin/env python3

### IMPORTS ###
import abc
import logging

### GLOBALS ###

### FUNCTIONS ###

### CLASSES ###
class LoggingService(metaclass = abc.ABCMeta):
    def __init__(self):
        self.logger = logging.getLogger(type(self).__name__)

    @abc.abstractmethod
    def debug(self, msg, *args, **kwargs):
        raise NotImplementedError

    @abc.abstractmethod
    def info(self, msg, *args, **kwargs):
        raise NotImplementedError

    @abc.abstractmethod
    def warning(self, msg, *args, **kwargs):
        raise NotImplementedError

    @abc.abstractmethod
    def error(self, msg, *args, **kwargs):
        raise NotImplementedError

    @abc.abstractmethod
    def critical(self, msg, *args, **kwargs):
        raise NotImplementedError

class InProcessLoggingService(LoggingService):
    def __init__(self):
        super().__init__()
        self.logger.debug("Inputs - None")
        self._data = {}

    def debug(self, msg, *args, **kwargs):
        self.logger.debug(msg, *args, **kwargs)

    def info(self, msg, *args, **kwargs):
        self.logger.info(msg, *args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        self.logger.warning(msg, *args, **kwargs)

    def error(self, msg, *args, **kwargs):
        self.logger.error(msg, *args, **kwargs)

    def critical(self, msg, *args, **kwargs):
        self.logger.critical(msg, *args, **kwargs)



#!/usr/bin/env python3

### IMPORTS ###

### GLOBALS ###

### FUNCTIONS ###

### CLASSES ###
class KneeDeepIOPluginsManagerException(Exception):
    pass

class ServiceAlreadyRegistered(KneeDeepIOPluginsManagerException):
    pass

class ServiceNotRegistered(KneeDeepIOPluginsManagerException):
    pass

class PluginAlreadyLoadedException(KneeDeepIOPluginsManagerException):
    pass

class PluginNotLoadedException(KneeDeepIOPluginsManagerException):
    pass



#!/usr/bin/env python3

### IMPORTS ###
from .configurationservice import ConfigurationService, InMemoryConfigurationService
from .loggingservice import LoggingService, InProcessLoggingService
#from .metricsreportingservice import MetricsReportingService
from .objectdatastoreservice import ObjectDatastoreService, InMemoryObjectDatastoreService, BasicFileBackedObjectDatastoreService

### GLOBALS ###

### FUNCTIONS ###

### CLASSES ###



#!/usr/bin/env python3

### IMPORTS ###
from .plugin import Plugin

### GLOBALS ###

### FUNCTIONS ###

### CLASSES ###
class ExampleLoggingPlugin(Plugin):
    required_services = ["logging"]
    def __init__(self, services):
        super().__init__(services)
        self.logger.debug("Inputs - services: %s", services)

    def setup(self):
        # This method should do whatever is needed to initialize the plugin's operation.
        self.logger.debug("setup method")

    def tick(self):
        # FIXME: Pick a better name for this
        # This method should do some small piece of work for the plugin.
        self.logger.debug("tick method")

    def teardown(self):
        # This method should do whatever is needed to un-setup the plugin, stopping operation.
        self.logger.debug("teardown method")


#!/usr/bin/env python3

### IMPORTS ###
import abc
import json
import logging

### GLOBALS ###

### FUNCTIONS ###

### CLASSES ###
class ObjectDatastoreService(metaclass = abc.ABCMeta):
    def __init__(self):
        self.logger = logging.getLogger(type(self).__name__)

    @abc.abstractmethod
    def get_value(self, key):
        raise NotImplementedError

    @abc.abstractmethod
    def set_value(self, key, value):
        raise NotImplementedError

class InMemoryObjectDatastoreService(ObjectDatastoreService):
    def __init__(self):
        super().__init__()
        self.logger.debug("Inputs - None")
        self._data = {}

    def get_value(self, key):
        self.logger.debug("Getting value for key: %s - %s", key, self._data[key])
        return self._data[key]

    def set_value(self, key, value):
        self.logger.debug("Setting value for key: %s - %s", key, value)
        self._data[key] = value

class BasicFileBackedObjectDatastoreService(ObjectDatastoreService):
    def __init__(self, filename):
        super().__init__()
        self.logger.debug("Inputs - filename: %s", filename)
        self._filename = filename
        self._data = {}

        self.logger.debug("Loading the contents of the file.")
        try:
            with open(self._filename, mode = 'r', encoding = 'utf8') as input_file:
                self._data = json.loads(input_file.read())
            self.logger.debug("After loading self._data: %s", self._data)
        except FileNotFoundError:
            self.logger.info("File not found when trying to load initial data.")

    def get_value(self, key):
        self.logger.debug("Getting value for key: %s - %s", key, self._data[key])
        return self._data[key]

    def set_value(self, key, value):
        self.logger.debug("Setting value for key: %s - %s", key, value)
        self._data[key] = value
        self.logger.debug("Saving the contents to the file.")
        with open(self._filename, mode = 'w', encoding='utf8') as output_file:
            output_file.write(json.dumps(self._data))



#!/usr/bin/env python3

# NOTE: This core module contains the loading and calling functionality.
from .pluginfactory import PluginFactory
from .exceptions import KneeDeepIOPluginsManagerException
from .exceptions import ServiceAlreadyRegistered
from .exceptions import ServiceNotRegistered
from .exceptions import PluginAlreadyLoadedException
from .exceptions import PluginNotLoadedException

### IMPORTS ###

### GLOBALS ###

### FUNCTIONS ###

### CLASSES ###
